from django.urls import path

from .views import styleguide

urlpatterns = [
    path("", styleguide, name="styleguide", kwargs={"page": "index"}),
    path("<str:page>/", styleguide, name="styleguide"),
]



import datetime
import json
import os
import pathlib
import random
from collections import namedtuple
from pathlib import Path
from typing import List, Literal
from uuid import uuid4

import requests
from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.files.storage import default_storage
from django.core.serializers.json import DjangoJSONEncoder
from django.utils import lorem_ipsum
from django.utils.text import capfirst
from wagtail.documents.models import Document
from wagtail.images import get_image_model

Image = get_image_model()


n = random.randint


def rbool(prob=0.5):
    return random.random() < prob


def lpar(num):
    """``num`` random paragraphs of lorem ipsum."""
    return "".join("<p>{0}</p>".format(p) for p in lorem_ipsum.paragraphs(num, False))


def lwords(num):
    """``num`` random lorem ipsum words."""
    return capfirst(lorem_ipsum.words(num, False))


def tel():
    """random international format (aus) telephone number eg. +61 (0) 412 345 678"""
    return "+{0} ({1}) {2} {3} {4}".format(
        n(10, 99), n(0, 9), n(100, 999), n(100, 999), n(100, 999)
    )


def rdate(future=True):
    """
    A random date. Set ``future`` to True for dates in the future, False for
    dates in the past. Possible dates span up to five years from today.
    """
    now = datetime.date.today()
    delta = datetime.timedelta(days=n(0, 1825))
    if future:
        return now + delta
    else:
        return now - delta


def r_heading(level, num):
    return "<h{0}>{1}</h{0}>".format(level, lwords(num))


def r_bold(num):
    return "<p><b>{0}</b></p>".format(lwords(num))


def r_italic(num):
    return "<p><i>{0}</i></p>".format(lwords(num))


def r_ordered_list(list_elements):
    list_elems = []
    for _ in range(list_elements):
        list_elems.append("<li>{0}</li>\n".format(lwords(n(2, 4))))
    return "<ol>\n{0}</ol>".format("\n".join(map(str, list_elems)))


def r_unordered_list(list_elements):
    list_elems = []
    for _ in range(list_elements):
        list_elems.append("<li>{0}</li>\n".format(lwords(n(2, 4))))
    return "<ul>\n{0}</ul>".format("\n".join(map(str, list_elems)))


def hr():
    return "<hr>"


def r_anchor(num):
    return "<a href='http://example.com'>{0}</a>".format(lwords(num))


def r_image():
    image = get_random_image()
    return '<embed alt="Example image" embedtype="image" format="fullwidth" id="{0}"/>'.format(
        image.id
    )


def embed_video():
    return '<embed embedtype="media" url="{0}"/>'.format(get_random_video_url())


def rich_text_example():
    """Layout all possible elements that can be used in rich text fields."""
    return open(os.path.join(os.path.dirname(__file__), "rich_text_example.html"), "r").read()


def random_rich_text_example(num, embeds=False):
    """
    Layout some possible elements that can be used in rich text fields in a random fashion
    Takes one argument, the number of elements. Set embeds=True if you wish to include images
    and videos
    """
    funcs = [
        lambda: r_heading(n(2, 5), n(1, 3)),
        lambda: r_bold(n(3, 5)),
        lambda: r_italic(n(3, 5)),
        lambda: r_ordered_list(n(3, 5)),
        lambda: r_unordered_list(n(3, 5)),
        lambda: r_anchor(n(3, 5)),
        lambda: lpar(n(3, 5)),
        hr,
    ]
    if embeds:
        funcs.extend([r_image, embed_video])
    return "\n".join([random.choice(funcs)() for _ in range(num)])


def json_dumps(x):
    """Dump a value to JSON using the Django JSON encoder"""
    return json.dumps(x, cls=DjangoJSONEncoder)


def shuf(l1):
    """Generate a shuffled copy of a list"""
    l2 = l1[:]
    random.shuffle(l2)
    return l2


class once(object):
    """A decorator that only calls the decorated function once"""

    has_run = False
    value = None
    fn = None

    def __init__(self, fn):
        self.fn = fn

    def __call__(self, *args, **kwargs):
        if self.has_run:
            return self.value

        self.value = self.fn(*args, **kwargs)
        self.has_run = True
        return self.value


def setup_images(test_img_dir=None, tag=None):
    def _is_image(filename):
        extensions = [".jpg", ".jpeg", ".png", ".gif", ".webp", ".svg"]
        return pathlib.Path(filename).suffix.lower() in extensions

    image_prefix = "original_images/test-images"

    if test_img_dir is None:
        HERE = os.path.join(os.path.dirname(__file__))
        test_img_dir = os.path.join(HERE, "test_images")
    CopiedImage = namedtuple("CopiedImage", ["name", "path"])
    copied_images: List[CopiedImage] = []
    # Copy all images from test_img_dir to assets directory
    for filename in os.listdir(test_img_dir):
        image_src_path = os.path.join(test_img_dir, filename)
        image_dest_path = os.path.join(image_prefix, filename)
        if not default_storage.exists(image_dest_path) and _is_image(filename):
            with open(image_src_path, "rb") as f:
                default_storage.save(image_dest_path, f)
        copied_images.append(CopiedImage(name=filename, path=image_dest_path))
    # Create Image models for each image that was found
    for i in copied_images:
        if _is_image(i.path):
            image_obj = Image.objects.create(title=i.name, file=i.path)
            if tag:
                image_obj.tags.add(tag)


def get_random_image(category: str = None) -> Image:
    images = Image.objects.order_by("?")
    if category:
        images = images.filter(tags__name__in=[category])
    return images.first()


def get_image(filename):
    return Image.objects.get(title=filename)


def setup_documents(test_document_dir=None):
    dest_dir = Path(settings.MEDIA_ROOT, "documents")

    if test_document_dir is None:
        here = Path(__file__).parent.absolute()
        test_document_dir = here.joinpath("test_documents")
    for src_file in test_document_dir.iterdir():
        doc_dest_path = dest_dir.joinpath(
            src_file.name
        ).as_posix()  # default_storage needs a string
        if not default_storage.exists(doc_dest_path) and src_file.is_file():
            with open(src_file.as_posix(), "rb") as doc_file:
                default_storage.save(doc_dest_path, doc_file)
    for file in dest_dir.iterdir():
        if file.is_file():
            Document.objects.create(title=file.name, file=file)


def get_random_doc():
    return Document.objects.order_by("?").first()


def get_random_video_url():
    return random.choice(
        [
            "https://youtu.be/H9VVkwRb_7M",
            "https://youtu.be/2WWwkArWajM",
            "https://youtu.be/yYaJWO_mNWs",
            "https://youtu.be/1kThvYlo2Lc",
            "https://youtu.be/Ji9qSuQapFY",
            "https://vimeo.com/155536745",
            "https://vimeo.com/179936131",
            "https://vimeo.com/188197250",
            "https://vimeo.com/175535505",
            "https://vimeo.com/188321226",
        ]
    )


def make_superuser():
    User = get_user_model()
    if User.USERNAME_FIELD == "email":
        User.objects.create_superuser(email="admin@admin.com", password="p")
    else:
        User.objects.create_superuser(username="admin", email="admin@admin.com", password="p")


def generate_form_streamfield(headings=True, file_field=True):
    """Generates one field of each possible type"""
    basic_types = ["text", "textarea", "email", "number", "date"]

    if headings:
        yield ("heading", "Basic Types")
    for field_type in basic_types:
        yield (
            "basic_field",
            {
                "label": field_type,
                "field_type": field_type,
                "help_text": lwords(10),
                "required": rbool(),
                "default_value": "",
                "placeholder": lwords(2),
            },
        )

    if headings:
        yield ("heading", "Choice Types")

    choice_types = ["select", "checkbox", "radio"]

    for field_type in choice_types:
        yield (
            "choice_field",
            {
                "label": field_type,
                "field_type": field_type,
                "choices": [lwords(n(1, 2)) for _ in range(n(2, 8))],
                "help_text": lwords(10),
                "required": rbool(),
            },
        )

    if file_field:
        if headings:
            yield ("heading", "File types")
        yield (
            "file_field",
            {
                "label": "File",
                "help_text": lwords(10),
                "required": rbool(),
            },
        )


def download_random_images(pic_type: Literal["people", "lorem"], limit=20):
    """
    Download some random images, if they exist, load them into the database.

    To use the images in factories pair with FuzzyImage(tag=pic_type)
    """
    path = Path(settings.MEDIA_ROOT, pic_type)
    if not path.exists():
        path.mkdir()
    # If directory is empty, create some test images
    if not list(path.iterdir()):
        for i in range(limit):
            if pic_type == "people":
                url = "https://thispersondoesnotexist.com"
            else:
                dimensions = [200, 300, 400, 500, 600]
                url = f"https://picsum.photos/{random.choice(dimensions)}/{random.choice(dimensions)}?random={i}"

            response = requests.get(url)
            image_file = path / f"{uuid4()}.jpg"
            with open(image_file, "wb") as f:
                f.write(response.content)

                try:
                    i = Image.objects.create(
                        file=f.name,
                    )
                    i.tags.add(pic_type)
                    print(f"Created image {image_file}")
                except Exception:
                    print("Image failed, deleting")
                    image_file.unlink()
    else:
        for image_file in path.iterdir():
            with open(image_file, "rb") as f:
                i = Image.objects.create(file=f.name)
                i.tags.add(pic_type)



import re
from typing import TYPE_CHECKING

from django.apps import apps
from django.conf import settings
from django.contrib import admin
from django.contrib.staticfiles.storage import staticfiles_storage
from django.http import Http404, HttpResponsePermanentRedirect
from django.shortcuts import render
from django.urls import include, path, re_path
from django.views.static import serve
from wagtail import urls as wagtail_urls
from wagtail.admin import urls as wagtailadmin_urls
from wagtail.contrib.sitemaps.views import sitemap
from wagtail.documents import urls as wagtaildocs_urls

from coop.styleguide import urls as styleguide_urls

if TYPE_CHECKING:
    from wagtail.contrib.settings.models import BaseSiteSetting
    from wagtail.images.models import AbstractImage


def asset_redirect(src, dest):
    try:
        asset_url = staticfiles_storage.url(dest)
    except ValueError:

        def view(request):
            raise Http404
    else:

        def view(request):
            return HttpResponsePermanentRedirect(asset_url)

    return re_path(r"^" + re.escape(src) + "$", view)


def favicon_redirect(request):
    if model_string := getattr(settings, "FAVICON_MODEL", None):
        try:
            app_name, model_name = model_string.split(".")
            Model: "BaseSiteSetting" = apps.get_model(app_name, model_name)
            instance = Model.for_request(request)
            if not instance:
                raise Http404("Favicon model not found")
            field = getattr(settings, "FAVICON_FIELD", "favicon")
            src: "AbstractImage" = getattr(instance, field)
            if not src:
                raise Http404("Favicon not found")
            dest = src.get_rendition("format-ico").url
            return HttpResponsePermanentRedirect(dest)
        except LookupError:
            raise Http404("Favicon settings not found")
    else:
        try:
            asset_url = staticfiles_storage.url("images/favicon.png")
        except ValueError:
            raise Http404("Favicon not found")
        return HttpResponsePermanentRedirect(asset_url)


def handler404(request, exception=None):
    request.is_preview = False
    return render(request, "layouts/404.html", {"exception": exception}, status=404)


def handler500(request):
    request.is_preview = False
    return render(request, "layouts/500.html", status=500)


urlpatterns = [
    path("favicon.ico", favicon_redirect),
    asset_redirect("humans.txt", "misc/humans.txt"),
    asset_redirect("robots.txt", "misc/robots.txt"),
    path("django-admin/", admin.site.urls),
    path("admin/", include(wagtailadmin_urls)),
    path("documents/", include(wagtaildocs_urls)),
    re_path(r"sitemap\.xml$", sitemap),
    path("_styleguide/", include(styleguide_urls)),
    path("404/", handler404),
    path("500/", handler500),
    path("", include(wagtail_urls)),
]

if settings.DEBUG or getattr(settings, "FORCE_ASSET_SERVING", False):

    def static(prefix, document_root):
        pattern = r"^%s(?P<path>.*)$" % re.escape(prefix.lstrip("/"))
        return [re_path(pattern, serve, kwargs={"document_root": document_root})]

    urlpatterns += static(settings.STATIC_URL, settings.STATIC_ROOT)
    urlpatterns += static(settings.MEDIA_URL, settings.MEDIA_ROOT)



import factory
from wagtail.blocks.stream_block import StreamValue
from wagtail_factories.blocks import StreamFieldFactory, StructBlockFactory

from coop.factories import FuzzyWords

from .blocks import ChoiceFieldBlock, FieldBlock, FileFieldBlock, FormBlocks


class BasicFieldBlockFactory(StructBlockFactory):
    label = FuzzyWords(3)
    help_text = FuzzyWords(5, True)
    required = factory.Faker("pybool")
    placeholder = FuzzyWords(3, True)

    class Meta:
        model = FieldBlock


class TextFieldBlockFactory(BasicFieldBlockFactory):
    field_type = "text"


class TextAreaBlockFactory(BasicFieldBlockFactory):
    field_type = "textarea"


class EmailFieldBlockFactory(BasicFieldBlockFactory):
    field_type = "email"


class NumberFieldBlockFactory(BasicFieldBlockFactory):
    field_type = "number"


class DateFieldBlockFactory(BasicFieldBlockFactory):
    field_type = "date"


class TimeFieldBlockFactory(BasicFieldBlockFactory):
    field_type = "time"


class CheckboxFieldFactory(BasicFieldBlockFactory):
    field_type = "checkbox"


class BaseChoiceFactory(StructBlockFactory):
    label = FuzzyWords(3)
    help_text = FuzzyWords(5, True)
    required = factory.Faker("pybool")
    choices = factory.Faker("texts", nb_texts=6, max_nb_chars=10)

    class Meta:
        model = ChoiceFieldBlock


class SelectChoiceFactory(BaseChoiceFactory):
    field_type = "select"


class CheckBoxChoiceFactory(BaseChoiceFactory):
    field_type = "checkbox"


class RadioChoiceFactory(BaseChoiceFactory):
    field_type = "radio"


class FileFieldFactory(StructBlockFactory):
    label = FuzzyWords(3)
    help_text = FuzzyWords(5, True)
    required = False

    class Meta:
        model = FileFieldBlock


class FormBlocksFactory(StreamFieldFactory):
    def __init__(self, **kwargs):
        factories = {}
        super().__init__(factories, **kwargs)

    def generate(self, instance, step):
        blocks = [
            ("basic_field", TextFieldBlockFactory()),
            ("basic_field", TextAreaBlockFactory()),
            ("basic_field", EmailFieldBlockFactory()),
            ("basic_field", NumberFieldBlockFactory()),
            ("basic_field", DateFieldBlockFactory()),
            ("basic_field", TimeFieldBlockFactory()),
            ("basic_field", CheckboxFieldFactory()),
            ("choice_field", SelectChoiceFactory()),
            ("choice_field", CheckBoxChoiceFactory()),
            ("choice_field", RadioChoiceFactory()),
            ("file_field", FileFieldFactory()),
            ("heading", "Heading"),
        ]

        return StreamValue(FormBlocks(), blocks)



from django.forms.fields import BooleanField, MultipleChoiceField


def form_data_to_text(form, html=False):
    """Helper method to turn a form response to a text (usually for email)"""
    data_string = ""

    for bound_field in form:
        if isinstance(bound_field.field, BooleanField):
            value = form.cleaned_data[bound_field.name]
            value_display = "Yes" if value else "No"
            if html:
                data_string += f"<p>{bound_field.label}: {value_display}</p>"
            else:
                data_string += f"\n\n{bound_field.label}: {value_display}"
        elif isinstance(bound_field.field, MultipleChoiceField):
            values = form.cleaned_data[bound_field.name]
            if html:
                data_string += f"<p>{bound_field.label}:"
            else:
                data_string += f"\n\n{bound_field.label}:"
            for choice in bound_field.field.choices:
                value_display = "Yes" if choice[0] in values else "No"
                if html:
                    data_string += f"<br/>{choice[0]} - {value_display}"
                else:
                    data_string += f"\n{choice[0]} - {value_display}"
            if html:
                data_string += "</p>"
        else:
            value = form.cleaned_data[bound_field.name]
            if not value:
                value = "Not filled out."
            if html:
                data_string += f"<p>{bound_field.label}:<br/>{value}</p>"
            else:
                data_string += f"\n\n{bound_field.label}:\n{value}"

    return data_string



"""
Remove the automatically generated site and home page
NOTE: This migration should be faked when converting a site to coop.
"""

from django.db import migrations


def remove_site(apps, schema):
    Site = apps.get_model('wagtailcore', 'site')
    Page = apps.get_model('wagtailcore', 'page')

    try:
        # Check if the site is premade from wagtailcore 0002_initial_data
        page = Page.objects.get(title="Welcome to your new Wagtail site!")

        site = Site.objects.get(root_page=page)
        site.delete()
        page.delete()
        # Fix the root page to show it has no children
        Page.objects.filter(pk=1).update(numchild=0)
    except Page.DoesNotExist:
        return


class Migration(migrations.Migration):

    dependencies = [
        ('wagtailcore', '__latest__'),
    ]

    operations = [
        migrations.RunPython(remove_site, migrations.RunPython.noop)
    ]



from django.utils.safestring import mark_safe
from wagtail import blocks
from wagtail.documents.blocks import DocumentChooserBlock


class UnwrappedStreamBlock(blocks.StreamBlock):
    """Removes the surrounding divs around streamblocks."""

    def render_basic(self, value, context=None):
        return mark_safe("\n".join(child.render(context=context) for child in value))


class LinkValue(blocks.StructValue):
    @property
    def href(self):
        if page := self.get("page"):
            return page.full_url
        elif doc := self.get("document"):
            return doc.url
        url = self.get("url", "")
        return url

    @property
    def text(self):
        if link_text := self.get("link_text"):
            return link_text
        elif page := self.get("page"):
            return page.title
        # Shouldn't get here
        return "Click here"


class BaseLinkBlock(blocks.StructBlock):
    def render_basic(self, value: LinkValue, context=None):
        return mark_safe(f'<a href="{value.href}">{value.text}</a>')


class InternalLink(BaseLinkBlock):
    link_text = blocks.CharBlock(required=False, help_text="Leave blank to use page name")
    page = blocks.PageChooserBlock()

    class Meta:
        value_class = LinkValue
        icon = "doc-empty"


class ExternalLink(BaseLinkBlock):
    link_text = blocks.CharBlock()
    url = blocks.URLBlock()

    class Meta:
        value_class = LinkValue
        icon = "fa-external-link"


class DownloadLink(BaseLinkBlock):
    link_text = blocks.CharBlock(help_text="e.g. Download File")
    document = DocumentChooserBlock()

    class Meta:
        value_class = LinkValue
        icon = "fa-download"


class LinkBlocks(UnwrappedStreamBlock):
    external = ExternalLink()
    internal = InternalLink()
    download = DownloadLink()

    class Meta:
        icon = "fa-list"



import json

import boto3
from django.conf import settings
from django_vite.core.asset_loader import DjangoViteAppClient, ManifestClient


class S3ManifestClient(ManifestClient):
    # This is a custom manifest client that loads the manifest from an S3/R2 bucket
    def load_manifest(self):
        s3 = boto3.client(
            "s3",
            aws_access_key_id=settings.AWS_S3_ACCESS_KEY_ID,
            aws_secret_access_key=settings.AWS_SECRET_ACCESS_KEY,
            endpoint_url=settings.AWS_S3_ENDPOINT_URL,
        )
        res = s3.get_object(Bucket=settings.AWS_STORAGE_BUCKET_NAME, Key="static/manifest.json")
        manifest_content = res["Body"].read()
        return json.loads(manifest_content)


class S3DjangoViteAppClient(DjangoViteAppClient):
    ManifestClient = S3ManifestClient



import jinja2.ext
from django.conf import settings
from django.contrib.staticfiles.storage import staticfiles_storage
from django_vite.core.asset_loader import DjangoViteAssetLoader
from django_vite.templatetags.django_vite import vite_asset, vite_hmr_client


def static_url(path, config="default"):
    if settings.DEBUG:
        asset_loader = DjangoViteAssetLoader.instance()
        client = asset_loader._get_app_client(config)
        if client.dev_mode:
            # Server running in debug and we're loading from the vite dev server, fetch from public
            return asset_loader.generate_vite_asset_url(path, config)
    return staticfiles_storage.url(path)


class Extension(jinja2.ext.Extension):
    def __init__(self, environment):
        super().__init__(environment)

        self.environment.globals.update(
            {
                "vite_hmr_client": vite_hmr_client,
                "vite_asset": vite_asset,
                "static": static_url,
            }
        )



from collections import OrderedDict

from django import forms
from django.utils.text import slugify
from wagtail import blocks

from coop.blocks import UnwrappedStreamBlock

from .forms import CustomForm


class BaseFieldBlock(blocks.StructBlock):
    def to_field(self, value):
        defaults = {
            "label": value["label"],
            "required": value["required"],
            "help_text": value["help_text"],
        }
        field_class = self.get_field_class(value, defaults)
        return slugify(value["label"]), field_class(**defaults)

    def get_context(self, value, parent_context=None):
        context = super().get_context(value, parent_context=parent_context)
        field_details = self.to_field(value)
        context.update(
            {
                "field_name": field_details[0],
                "field": field_details[1],
                "bound_field": parent_context["form"][slugify(value["label"])],
            }
        )
        return context

    class Meta:
        abstract = True


class FieldBlock(BaseFieldBlock):
    label = blocks.CharBlock()
    field_type = blocks.ChoiceBlock(
        choices=[
            ("text", "Single line text"),
            ("textarea", "Multi line text"),
            ("email", "Email"),
            ("number", "Number"),
            ("date", "Date"),
            ("time", "Time"),
            ("checkbox", "Checkbox"),
        ]
    )
    help_text = blocks.CharBlock(required=False)
    required = blocks.BooleanBlock(default=True, required=False)
    placeholder = blocks.CharBlock(required=False)
    default_value = blocks.CharBlock(required=False)

    def get_field_class(self, value, defaults):
        defaults.update({"initial": value["default_value"]})
        field_type = value["field_type"]
        widget_props = {"attrs": {"placeholder": value["placeholder"]}}
        if field_type == "textarea":
            defaults.update({"widget": forms.Textarea(**widget_props)})
            return forms.CharField
        elif field_type == "email":
            defaults.update({"widget": forms.EmailInput(**widget_props)})
            return forms.EmailField
        elif field_type == "number":
            defaults.update({"widget": forms.NumberInput(**widget_props)})
            return forms.IntegerField
        elif field_type == "date":
            widget_props["attrs"]["type"] = "date"
            defaults.update({"widget": forms.DateInput(**widget_props)})
            return forms.DateField
        elif field_type == "time":
            widget_props["attrs"]["type"] = "time"
            defaults.update({"widget": forms.TimeInput(**widget_props)})
            return forms.TimeField
        elif field_type == "checkbox":
            return forms.BooleanField
        # Default - assume text field
        defaults.update({"widget": forms.TextInput(**widget_props)})
        return forms.CharField

    class Meta:
        icon = "fa-edit"
        template = "blocks/form/form_field.html"


class ChoiceFieldBlock(BaseFieldBlock):
    label = blocks.CharBlock()
    field_type = blocks.ChoiceBlock(
        choices=[
            ("select", "Dropdown"),
            ("checkbox", "Checkboxes"),
            ("radio", "Radio buttons"),
        ]
    )
    choices = blocks.ListBlock(blocks.CharBlock(icon="fa-square-o"))
    help_text = blocks.CharBlock(required=False)
    required = blocks.BooleanBlock(default=True, required=False)

    def get_field_class(self, value, defaults):
        widget = {
            "select": forms.widgets.Select(),
            "checkbox": forms.widgets.CheckboxSelectMultiple(),
            "radio": forms.widgets.RadioSelect(),
        }[value["field_type"]]
        defaults.update({"choices": [(c, c) for c in value["choices"]], "widget": widget})
        if value["field_type"] == "checkbox":
            return forms.MultipleChoiceField
        return forms.ChoiceField

    class Meta:
        icon = "fa-check-square-o"
        template = "blocks/form/form_field.html"


class FileFieldBlock(BaseFieldBlock):
    label = blocks.CharBlock()
    help_text = blocks.CharBlock(required=False)
    required = blocks.BooleanBlock(default=True, required=False)

    def get_field_class(self, value, defaults):
        return forms.FileField

    class Meta:
        icon = "fa-file"
        template = "blocks/form/file_upload.html"


class HeadingBlock(blocks.CharBlock):
    class Meta:
        template = "blocks/form/form_heading.html"
        icon = "fa-header"


class FormBlocks(UnwrappedStreamBlock):
    basic_field = FieldBlock()
    choice_field = ChoiceFieldBlock()
    file_field = FileFieldBlock()
    heading = HeadingBlock()

    def get_fields_from_blocks(self, blocks):
        for child in blocks:
            block = child.block
            if hasattr(block, "to_field"):
                yield block.to_field(child.value)

    def get_form(self, blocks, **kwargs):
        extra_fields = OrderedDict(self.get_fields_from_blocks(blocks))
        return CustomForm(extra_fields, **kwargs)

    class Meta:
        icon = "fa-pencil"


class FormStreamBlock(UnwrappedStreamBlock):
    form = FormBlocks()

    class Meta:
        block_counts = {
            "form": {
                "max_num": 1,
            }
        }
